/**
 * Controller interfaces for hdparm.
 *
 * This file allows HDIO_ ioctl operations to be performed on different
 * controllers which provide alternative interfaces. Usually this is
 * required when a RAID controller provides additional support for accessing
 * information in the underlying devices.
 *
 * The way in which the controllers are described is similar to that used
 * by smartmontools, but is implemented independantly.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/hdreg.h>

#include "sgio.h"
#include "controller.h"

extern int verbose;

/** Maximum device name length */
#define DEVICE_NAME_LEN (128)

/** ioctl interface for HighPoint RAID controllers. */
#define HPTIO_CTL                       0x03ff

/** Structure used for all HPTIO_CTL operations */
typedef struct highpoint_device_op {
    int32_t controller;
    int32_t channel;
    int32_t hdio_op; /* HDIO_ ioctl operation; not all are supported */
    int32_t device;
    /* Followed by a variable sized bufffer for the data */
} highpoint_device_op;

/**
 * Structure describing the device to target on the HighPoint controllers.
 */
typedef struct highpoint_deviceid {
    int32_t controller;
    int32_t channel;
    int32_t device; /* Port-multiplier device, if any; 0 if none */
} highpoint_deviceid;



/**
 * Perform an ioctl operation on a highpoint controller, rather
 * than directly to the device.
 *
 * Actually we only support the HDIO_* ioctl operations.
 *
 * @param[in]     fd        File descriptor on which to operate
 * @param[in]     target    Highpoint device to target
 * @param[in]     cmd       ioctl command to use
 * @param[in,out] arg       ioctl buffer
 * @param[in]     arglen    length of the ioctl buffer
 *
 * @retval 0    success
 * @retval -ve  failure
 */
static int highpoint_ioctl(int                fd,
                           highpoint_deviceid target,
                           int                cmd,
                           char              *arg,
                           size_t             arglen)
{
    char hpt_cmdbuffer[sizeof(highpoint_device_op) + arglen];
    highpoint_device_op *hpt_cmd = (highpoint_device_op *)&hpt_cmdbuffer[0];
    int retval;
    int outlen = arglen;

    /* Report anything that has a zero length as an invalid operation;
     * the highpoint controller's operations do not support the flag/integer
     * operations, where 'arg' is a raw value rather than a pointer to a
     * value.
     */
    if (arglen == 0)
    {
        errno = EINVAL;
        return -1;
    }

    /* We can only process ioctl operations that are for HDIO_* */
    if ( (cmd & 0xFFFFFF00) != 0x0300 )
    {
        errno = EINVAL;
        return -1;
    }

    hpt_cmd->controller = target.controller;
    hpt_cmd->channel    = target.channel;
    hpt_cmd->device     = target.device;
    hpt_cmd->hdio_op    = cmd;

    /* Append the data to the buffer */
    memcpy( &hpt_cmdbuffer[sizeof(*hpt_cmd)], arg, arglen);

    if (verbose)
    {
        fprintf(stderr,
                "HighPoint(%i/%i/%i) ioctl: 0x%x\n",
                target.controller,
                target.channel,
                target.device,
                cmd);
        if (cmd == HDIO_DRIVE_CMD)
        {
            fprintf(stderr,
                    "  DRIVE_CMD: 0x%x, feature=0x%x, nsector=0x%x\n",
                    ((unsigned char*)arg)[0],
                    ((unsigned char*)arg)[2],
                    ((unsigned char*)arg)[3]);

            if (((unsigned char*)arg)[0] == ATA_OP_CHECKPOWERMODE1)
            {
                outlen = 4;
            }
        }
    }

    retval = ioctl(fd, HPTIO_CTL, hpt_cmd);
    if (retval == 0)
    {
        /* On success, copy the data from the temporary buffer to
         * the output buffer.
         */
        if (verbose)
            fprintf(stderr, "Copying out %i bytes\n", outlen);
        memcpy( arg, &hpt_cmdbuffer[sizeof(*hpt_cmd)], outlen);
    }
    if (verbose)
    {
        fprintf(stderr, "  returned %i\n", retval);
    }
    return retval;
}



/** Controller type for ioctl operations */
static enum controller_type {
    ct_normal,
    ct_highpoint
} controller_type;

/** Information about the highpoint device, if one is selected */
static highpoint_deviceid controller_highpoint = { 0,0,0 };


/**
 * A generalised dispatch for ioctl operations, which calls either
 * the standard ioctl, or through a special controller interface for certain
 * hardware.
 *
 * Due to the nature of the operations, we 'know' about the length
 * of certain structures.
 *
 * @param[in]     fd        File descriptor on which to operate
 * @param[in]     cmd       ioctl command to use
 * @param[in,out] arg       ioctl buffer
 *
 * @retval 0    success
 * @retval -ve  failure
 */
int controller_ioctl(int                fd,
                     int                cmd,
                     ...)
{
    va_list ap;
    int     arglen = 0; /* initially unknown */
    void   *arg;

    va_start(ap, cmd);
    arg = va_arg(ap, void *);
    va_end(ap);

    switch (cmd)
    {
        case HDIO_GET_IDENTITY:
            arglen = 512;
            break;

        case HDIO_DRIVE_CMD:
            {
                int command = ((unsigned char *)arg)[0];
                int nsector = ((unsigned char *)arg)[3];
                arglen = 4+512 * nsector;
            }
            break;
    }

    if (verbose)
    {
        fprintf(stderr, "ioctl(ctrl=%i): 0x%x\n", controller_type, cmd);
    }
    switch (controller_type)
    {
        case ct_normal:
            return ioctl(fd, cmd, arg);

        case ct_highpoint:
            return highpoint_ioctl(fd,
                                   controller_highpoint,
                                   cmd,
                                   arg,
                                   arglen);
    }

    errno = EINVAL;

    /* No support for this controller type (should not happen) */
    return -1;
}


/**
 * Parse a controller name, for use in subsequent operations on the
 * device. The controller is only expected to be required when operations
 * are performed on hardware underlying a device which does not directly
 * expose them - typically RAID controllers.
 *
 * @param[in] arg   Controller label:
 *                      'normal'    normal device operations
 *                      'hpt,L/M/N' HighPoint controllers, where
 *                                      L is the controller
 *                                      M is the channel
 *                                      N is the port-multiplier device
 *                                  N may be omitted.
 *
 * @note Help will be output if no controller label is understood,
 *       and the command will exit.
 */
void parse_controller(const char *arg)
{
    int help = 0;

    if (strcasecmp(arg, "help") == 0)
    {
        help = 1;
    }
    if (strcasecmp(arg, "normal") == 0)
    {
        controller_type = ct_normal;
        return;
    }
    if (strncasecmp(arg, "hpt,", 4) == 0)
    {
        int slash = 0;

        /* Skip 'hpt,' */
        arg += 4;

        controller_type = ct_highpoint;
        controller_highpoint.controller = 0;
        controller_highpoint.channel = 0;
        controller_highpoint.device = 0;

        /* Parsing code is intentionally similar to that in smartmontools,
         * but is implemented slightly differently.
         */
        while (*arg)
        {
            if (*arg == '/')
            {
                slash++;
                if (slash == 3)
                {
                   fprintf(stderr, "Option '--controller hpt,L/M/N' supports 2-3 items\n");
                   exit(1);
                }
                arg++;
            }
            else if (isdigit(*arg))
            {
                char *end;
                long val = strtol(arg, &end, 10);
                if (val <= 0)
                {
                    fprintf(stderr, "Option '--controller hpt,L/M/N' could not be parsed (at '%s')\n", arg);
                    exit(1);
                }
                switch (slash)
                {
                    case 0:
                        controller_highpoint.controller = (int32_t)val;
                        break;
                    case 1:
                        controller_highpoint.channel = (int32_t)val;
                        break;
                    case 2:
                        controller_highpoint.device = (int32_t)val;
                        break;
                }
                arg = end;
            }
            else
            {
                fprintf(stderr, "Option '--controller hpt,L/M/N' could not be parsed (at '%s')\n", arg);
                exit(1);
            }
        }

        if (slash == 1)
        {
           /* Only 2 components were given, so default the device to
            * 1
            */
           controller_highpoint.device = 1;
        }
        return;
    }

    /* If we reached here, we need to print out the help message, either
     * because it was explicitly requested, or because the controller was
     * not understood.
     */
    if (!help)
    {
        fprintf(stderr, "Controller type '%s' not recognised\n\n", arg);
    }

    fprintf(stderr, "Controller types:\n"
                    "    normal         Standard ioctl operations\n"
                    "    hpt,L/M/N      HighPoint controllers\n");

    exit(help ? 0 : 1);
}


/**
 * Returns a description of the device accessed through the controller,
 * in a static buffer.
 *
 * The function is intended to be used to display alongside the device
 * name to make a readable string.
 *
 * @return pointer to a zero-terminated device name, in a static buffer
 */
const char *controller_devicename(void)
{
    static char buffer[DEVICE_NAME_LEN];

    switch (controller_type)
    {
        case ct_normal:
            strcpy(buffer, "");
            break;

        case ct_highpoint:
            snprintf(buffer, sizeof(buffer),
                     "HighPoint device %i/%i/%i",
                     controller_highpoint.controller,
                     controller_highpoint.channel,
                     controller_highpoint.device);
            break;

        default:
            strcpy(buffer, "unknown device");
            break;
    }

    return buffer;
}

/* EOF */
