/**
 * Controller interfaces for hdparm.
 *
 * This file allows HDIO_ ioctl operations to be performed on different
 * controllers which provide alternative interfaces. Usually this is
 * required when a RAID controller provides additional support for accessing
 * information in the underlying devices.
 *
 * The way in which the controllers are described is similar to that used
 * by smartmontools, but is implemented independantly.
 */

#ifndef CONTROLLER_H
#define CONTROLLER_H


/**
 * A generalised dispatch for ioctl operations, which calls either
 * the standard ioctl, or through a special controller interface for certain
 * hardware.
 *
 * Due to the nature of the operations, we 'know' about the length
 * of certain structures.
 *
 * @param[in]     fd        File descriptor on which to operate
 * @param[in]     cmd       ioctl command to use
 * @param[in,out] arg       ioctl buffer
 *
 * @retval 0    success
 * @retval -ve  failure
 */
int controller_ioctl(int                fd,
                     int                cmd,
                     ...);


/**
 * Parse a controller name, for use in subsequent operations on the
 * device. The controller is only expected to be required when operations
 * are performed on hardware underlying a device which does not directly
 * expose them - typically RAID controllers.
 *
 * @param[in] arg   Controller label:
 *                      'normal'    normal device operations
 *                      'hpt,L/M/N' HighPoint controllers, where
 *                                      L is the controller
 *                                      M is the channel
 *                                      N is the port-multiplier device
 *                                  N may be omitted.
 *
 * @note Help will be output if no controller label is understood,
 *       and the command will exit.
 */
void parse_controller(const char *arg);


/**
 * Check whether the controller support SGIO operations.
 *
 * For RAID devices, the controller may support SGIO operations on the
 * device itself, and in that case - when the controller state is
 * 'normal', we will return results. However, if they have explicitly
 * requested a controller device, that means that they wish to talk to
 * the underlying device. In that case, we deny the SGIO, which means
 * that the operation will fall back to a HDIO_* operation which will
 * use the necessary pass-through parameters.
 *
 * For example, a RAID controller might present its view of the SMART
 * data if queried directly (ie in controller='normal' state). But if
 * we want to talk to the underlying device directly, and get its
 * information, we could specify the controller and device details,
 * and will obtain them separately.
 *
 * This may be a little HighPoint-specific, so may need to be considered
 * for refactoring based on how other controllers work.
 *
 * @retval 1 controller supports SGIO
 * @retval 0 controller should not have SGIO operations performed
 */
int controller_supports_sgio(void);


/**
 * Returns a description of the device accessed through the controller,
 * in a static buffer.
 *
 * The function is intended to be used to display alongside the device
 * name to make a readable string.
 *
 * @return pointer to a zero-terminated device name, in a static buffer
 */
const char *controller_devicename(void);

#endif